#!/usr/bin/env python3

class LineTokenizer:
    "Make a collection of lines available either as lines or tokens."
    def __init__(self, lines):
        self.lines = lines
        self.pretokenizer = None
        self.token_index = 0
        self.tokens = []
        self.tokenize()
    def popline(self):
        "Grab the next line and make it the token buffer."
        if not self.lines:
            #sys.stderr.write("Popline returns None\n")
            return None
        else:
            #sys.stderr.write("Popline starts with: %s\n" % self)
            res = self.lines[0]
            self.lines.pop(0)
            self.tokens = []
            if self.lines:
                self.tokenize(self.pretokenizer)
            #sys.stderr.write("In popline, I return %s: %s\n" % (`res`, self))
            return res
    def pushline(self, line):
        "Replace the token buffer with the current line."
        self.lines = [self.line] + self.lines
        self.tokenize(self.pretokenizer)
        #sys.stderr.write("Pushline leaves: %s\n" % self)
    def peekline(self):
        "Return the token buffer"
        if not self.lines:
            return None
        else:
            return self.lines[0]
    def tokenize(self, new_pretokenizer=None):
        "Split a line on tabs and whitespaces, but not linefeeds."
        self.pretokenizer = new_pretokenizer
        if self.lines:
            if self.pretokenizer:
                line = self.pretokenizer(self.lines[0])
            else:
                line = self.lines[0]
            self.tokens = line.split()
            #sys.stderr.write("In tokenize, I split: " + `self` + '\n')
    def restore_newlines(self):
        self.tokens.append("\n")
        for i in range(len(self.lines)):
            self.lines[i] += "\n"
    def token_pop(self, count=1):
        "Get a token."
        if not self.lines:
            return None
        #sys.stderr.write("In token_pop, I see: " + `self` + '\n')
        res = self.tokens[0]
        self.tokens = self.tokens[count:]
        if not self.tokens:
            if not self.lines:
                #sys.stderr.write("In token_pop, I return None: " + `self` + '\n')
                return None
            self.popline()
        self.token_index += 1
        #sys.stderr.write("In token_pop, I return: " + `self` + '\n')
        return res
    def token_peek(self):
        "Peek at the next token."
        if self.tokens:
            return self.tokens[0]
        else:
            return None		# list empty means we're out of data

    def token_push(self, tok):
        "Push back a token."
        self.tokens = [tok] + self.tokens
        # We do *not* alter the source line!
        self.token_index -= 1
    def __str__(self):
        "Display the state of the object."
        return "<tokens=%s, lines=%s>" % (self.tokens, self.lines)
    __repr__ = __str__

import exceptions

class CDeclarationError(exceptions.Exception):
    def __init__(self, message, retval=1):
        self.message = message
        self.retval = retval

class CDeclarationParser:
    "Parse a C declaration,"
    def __init__(self, io, handler, notifier=lambda x: None):
        self.io = io
        self.handler = handler
        self.notifier = notifier
        self.tokencount = 0
        self.state_stack = []
        self.construct_stack = []
        self.typedefs = {}
        self.declaration()
    def token_peek(self):
        return self.io.token_peek()	# FIXME: skip C comments
    def token_pop(self):
        self.tokencount += 1
        return self.io.token_pop()	# FIXME: skip C comments
    def checkin(self, label, opt):
        print " " * len(self.construct_stack) + "->" + label + " " + ("required", "optional")[opt]
        self.state_stack.append(self.tokencount)
        self.construct_stack.append(label)
    def checkout(self, label, opt):
        nonempty = self.state_stack[-1] < self.tokencount
        self.state_stack.pop()
        self.construct_stack.pop()
        if not opt and not nonempty:
            raise CDeclarationError("missing required element in " + label)
        else:
            print " " * len(self.construct_stack) + "<-" + label + " " + ("not found", "found")[nonempty]
            return nonempty
    def expect(self, *args):
        yes = self.token_peek() in args
        print " " * len(self.construct_stack) + "expect" + `args` + " = " + "ny"[yes] + " (" + self.token_peek() + ")"
        if yes:
            # One of only two places we pop the token stack
            print "Popped terminal:", self.token_pop()
        return yes
    def identifier(self, opt):
        self.checkin("identifier", opt)
        id = self.token_peek()
        if not id or (not id[0].isalpha and isalpha[0] != "_"):
            raise CDeclarationError("saw %s where identifier expected " % id)
        else:
            # And here is the other.
            self.handler(id, self.construct_stack)
            self.token_pop()
        self.checkout("identifier", opt)
    # Rest is hand-compiled from "A.13 of the C Programming Language", 2nd ed.
    # It covers the entire declaration syntax, except for initializers.
    # It does not cover function definitions.  There are two things we
    # don't do quite right because they can't be done LL(1), see below.
    def declaration(self):
        self.declaration_specifiers(opt=0)
        self.init_declarator_list(opt=0)	# Technically, is opt
        self.expect(";")
    def declaration_list(self, opt):
        self.checkin("declaration_list", opt)
        self.declaration_specifiers(opt=0)
        while self.declaration_specifiers(opt=1):
            continue
        return self.checkout("declaration_list", opt)
    def declaration_specifiers(self, opt):
        self.checkin("declaration_specifiers", opt)
        if \
        self.storage_class_specifier(opt=1) or \
        self.type_specifier(opt=1) or \
        self.type_qualifier(opt=1):
            self.declaration_specifiers(opt=1)
        return self.checkout("declaration_specifiers", opt)
    def storage_class_specifier(self, opt):
        self.checkin("storage_class_specifier", opt)
        self.expect("static", "extern", "typedef")	# auto, register
        return self.checkout("storage_class_specifier", opt)
    def type_specifier(self, opt):
        self.checkin("type_specifier", opt)
        self.expect("void", "char", "short", "int", "long", "float", "double", "signed", "unsigned") \
        or \
        self.struct_or_union_specifier(opt) \
        or \
        self.identifier(opt)
        # We've left out one legal alternative, an enum, because we
        # never expect to encounter it on a man page.
        # We accept any identifier here because we can't know all typedefs
        # in advance -- that would require parsing include files and a
        # semi-infinite amount of hair.
        return self.checkout("type_specifier", opt)
    def type_qualifier(self, opt):
        self.checkin("type_qualifier", opt)
        self.expect("const", "volatile")
        return self.checkout("type_qualifier", opt)
    def struct_or_union_specifier(self, opt):
        self.checkin("struct_or_union_specifier", opt)
        if not (self.expect("struct","union") and self.identifier(opt=0)):
            self.expect("{") and \
            self.struct_declaration_list(opt=0) and \
            self.expect("}")
        return self.checkout("struct_or_union_specifier", opt)
    def struct_declaration_list(self, opt):
        self.checkin("struct_declaration_list", opt)
        self.struct_declaration(opt=0)
        while self.struct_declaration(opt=1):
            continue
        return self.checkout("struct_declaration_list", opt)
    def init_declarator_list(self, opt):
        self.checkin("init_declarator_list", opt)
        # This is where we lop off the initializer branch.
        self.declarator(opt=0)
        while self.expect(","):
            self.init_declaration(opt=1)
        return self.checkout("init_declarator_list", opt)
    def struct_declaration(self, opt):
        self.checkin("struct_declaration", opt)
        self.specifier_qualifier_list(opt=0)
        self.struct_declarator_list(opt=0)
        return self.checkout("struct_declaration", opt)
    def specifier_qualifier_list(self, opt):
        self.checkin("specifier_qualifier_list", opt)
        self.type_specifier(opt=0) or self.type_qualifier(opt=0)
        while self.type_specifier(opt=1) or self.type_qualifier(opt=1):
            continue
        self.specifier_qualifier_list(opt=1)
        return self.checkout("specifier_qualifier_list", opt)
    def struct_declarator_list(self, opt):
        self.checkin("struct_declarator_list", opt)
        self.struct_declarator(opt=0)
        while self.struct_declaration(opt=1):
            continue
        return self.checkout("struct_declarator_list", opt)
    def struct_declarator(self, opt):
        self.checkin("struct_declarator", opt)
        self.declarator(opt=1)
        if self.expect(":"):
            self_constant_expression(opt=0)
        return self.checkout("struct_declarator", opt)
    def declarator(self, opt):
        self.checkin("declarator", opt)
        self.pointer(opt=1)
        self.direct_declarator(opt=0)
        return self.checkout("declarator", opt)
    def direct_declarator(self, opt):
        self.checkin("direct_declarator", opt)
        # This is not quite the A.13 production, which is
        # direct-declarator:
        #     identifer
        #     ( declarator)
        #     direct-declarator [ constant-expression? ]...
        #     direct-declarator ( parameter-type-list )
        #     direct-declarator ( identifier-list )
        # This be parsed LL(1) the way we're doing.  We implement this:
        # direct-declarator-prefix:
        #     identifier
        #     ( declarator )
        # direct-declarator:
        #     direct-declarator-prefix [ constant-expression? ]...
        #     direct-declarator-prefix ( parameter-type-list )
        #     direct-declarator-prefix ( identifier-list )
        # This won't catch weird shit like foo(bar)[MUMBLE].
        # Let's hope it doesn't miss any real-world cases.
        if self.expect("("):
            self.declarator(opt=0)
            self.expect(")")
        else:
            self.identifier(opt=0)
        if self.expect("("):
            self.parameter_type_list(opt=0) or self.identifier_list(1)
            self.expect(")")
        else:
            while self.expect("["):
                self.constant_expression(opt=1)
                self.expect("]")
        return self.checkout("direct_declarator", opt)
    def pointer(self, opt):
        self.checkin("pointer", opt)
        self.expect("*")
        if self.type_qualifier_list(opt=1):
            self.pointer(opt=1)
        return self.checkout("pointer", opt)
    def type_qualifier_list(self, opt):
        self.checkin("type_qualifier_list", opt)
        while self.type_qualifier(opt=1):
            continue
        return self.checkout("type_qualifier_list", opt)
    def parameter_type_list(self, opt):
        self.checkin("parameter_type_list", opt)
        self.parameter_list(opt=0)
        if self.expect(","):
            self.expect("...")
        return self.checkout("parameter_type_list", opt)
    def parameter_list(self, opt):
        self.checkin("parameter_list", opt)
        self.parameter_declaration(opt=0)
        while self.expect(","):
            self.parameter_declaration(opt=0)
        return self.checkout("parameter_list", opt)
    def parameter_declaration(self, opt):
        self.checkin("parameter_declaration", opt)
        self.declaration_specifiers(opt=0)
        self.declarator(opt=0) \
        or \
        self.abstract_declarator(opt=1)
        return self.checkout("parameter_declaration", opt)
    def identifier_list(self, opt):
        self.checkin("identifier_list", opt)
        self.identifier(opt=0)
        while self.expect(","):
            self.identifier(opt=0)
        return self.checkout("identifier_list", opt)
    def type_name(self, opt):
        self.checkin("type_name", opt)
        self.specifier_qualifier_list(opt=0)
        self.abstract_declarator(opt=1)
        return self.checkout("type_name", opt)
    def abstract_declarator(self, opt):
        self.checkin("abstract_declarator", opt)
        if self.pointer(opt=1):
            self.direct_abstract_declarator(opt=1)
        else:
            self.direct_abstract_declarator(opt=0)
        return self.checkout("abstract_declarator", opt)
    def direct_abstract_declarator(self, opt):
        self.checkin("direct_abstract_declarator", opt)
        # The actual production is
        # direct_abstract_declarator:
        #    ( abstract-declarator )
        #    direct_abstract_declarator? [ constant-expression? ]
        #    direct_abstract_declarator? ( parameter-type-list? )
        # This too cannot be parsed LL(1)  So we do this:
        # direct_abstract_declarator-prefix:
        #    ( abstract_declarator ) [ constant-expression? ]...
        #    ( abstract_declarator ) ( parameter-type-list? )
        # and hope it's good enough to catch the real-world cases.
        if self.expect("("):
            self.abstract_declarator(opt=0)
            self.expect(")")        
        if self.expect("("):
            self.parameter_type_list(opt=0)
            self.expect(")")
        else:
            while self.expect("["):
                self.constant_expression(opt=1)
                self.expect("]")
        return self.checkout("direct_abstract_declarator", opt)
    def typedef_name(self, opt):
        self.checkin("typedef_name", opt)
        self.identifier(opt=0)
        return self.checkout("typedef_name", opt)
    def constant_expression(self, opt):
        self.checkin("constant_expression", opt)
        # Full grammar has expressions here.
        # All we're ever going to see is constants, thank goodness.
        self.identifier(opt=1)
        return self.checkout("constant_expression", opt)

if __name__ == "__main__":
    import sys
    def post(token, stack):
        print "I see %s with %s" % (token, stack)
    def notify(msg):
        print msg
    io = LineTokenizer(sys.stdin.readlines())
    print io
    try:
        CDeclarationParser(io, post, notify)
    except CDeclarationError, e:
        print e.message




#!/usr/bin/env python3
# Generate the bug summary page
# All the distro-dependent stuff is imported from buglist.

import string, buglist, time

bugs = buglist.Buglist()

old_accepted = 594	# Represents patches formerly tagged 'y'.

applied = len(bugs.pagelist(include="npr"))
pending = len(bugs.pagelist(include="n*"))
accepted = old_accepted + len(bugs.pagelist(include="yp"))
rejected = len(bugs.pagelist(include="r"))

total = sum(bugs.counts)
already =  bugs.counts[7]
errors =  sum(bugs.counts[1:7])	# Types 1-6 	

ok = (total - already) - errors - applied

def percent(x):
    return x * 100.0 / (total - already)

print '''<?xml version="1.0" encoding="ISO-8859-1"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<link rev="made" href="mailto:esr@snark.thyrsus.com" />
<link rel="stylesheet" href="/~esr/sitestyle.css" type="text/css" />
<meta name="description" content="" />
<meta name="keywords" content="" />
<meta name="MSSmartTagsPreventParsing" content="TRUE" />
<title>Manual page glitches</title>
</head>
<body>

<div id="Header">
<table width="100%%" cellpadding="0" summary="Canned page header">
<tr>
<td>Manual page glitches</td>
<td align="right">%s</td>
</tr>
</table>
</div>

<div id="Menu">
	<hr/>
	<a href="/~esr" title="My home page">Home Page</a><br />
	<a href="/~esr/whatsnew.html" title="New on this site">What is New</a><br />
	<a href="/~esr/sitemap.html" title="Map of the site">Site Map</a><br />
	<a href="/~esr/software.html" title="Software I maintain">Software</a><br />
	<a href="/~esr/projects.html" title="My projects">Projects</a><br />
	<a href="/~esr/faqs/" title="My FAQ documents">HOWTOs</a><br />
	<a href="/~esr/writings/" title="Essays and ruminations">Essays</a><br />
	<a href="/~esr/personal.html" title="Portrait of the author">Personal</a><br />
	<a href="http://www.ibiblio.org/esrblog/">Weblog</a><br/>
        <a href="/~esr/netfreedom/">Freedom!</a><br />
	<a href="/~esr/guns/">Firearms!</a><br />
	<hr/>
    <p>
      <a href="http://validator.w3.org/check/referer"><img
          src="http://www.w3.org/Icons/valid-xhtml10"
          alt="Valid XHTML 1.0!" height="31" width="88" /></a>
    </p>

</div>

<div id="Content">

<p>I maintain a man-page-to-DocBook converter, <a
href="index.html">doclifter</a>. A side effect of this program is that
it serves as a validator for the correctness and portability of the
markup used on Unix manual pages.  I test it by running it against all
the manual pages in a full %s; there are %d of these on my development
machine, of which %d already have DocBook masters.  It converts %d
(%02.2f%%) of the remaining %d into valid XML-DocBook.</p>

<p>Most of the remaining %02.2f%% of errors happen because groff(1)
and its kin have weak-to-nonexistent validity checking.  Often,
doclifter fails because of outright errors in macro usage that groff
does not catch.  Sometime it fails on constructions that are legal but
perverse.  Very occasionally it throws an error because a man page is
correct but has a structure that cannot be translated to DocBook.  I
keep a database of patches for such problems, and periodically
try to push fix patches out to the manual-page maintainers.</p>

<p>(These are lower numbers and a higher error rate than in some
previous reports because I now use i3 rather than GNOME or KDE. Many
of the userland manuals that I used to check are no longer installed
where my test procedure can see them. Because bad markup tends to be
concentrated in the older manual pages of core tools, a larger random
sample pulls down the error rate.)

<p>Even if you do not care about DocBook, this cleanup work benefits
all third-party manual page viewers, including the GNOME and KDE
documentation browsers; groff constructions that confuse doclifter
are very likely to produce visible problems on these.</p>

<p>The table below is a listing of the %d (%02.2f%%) pages on which
doclifter fails, but the failure can be prevented with a fix patch to
the manual page source.  %d pages (%02.2f%%) remain intractable,
generally due to markup problems more severe than a point patch can
address.  I am working with the individual projects responsible to get
those cleaned up.</p>

<p>It is likely that you are reading this because you have received
email telling you that patches are associated with your name or list
address.  Please consider incorporating them, or equivalents, in your
next release.  Also, please write back and tell me what you plan to do
so I can keep my database up-to-date.</p>

<p>If you are not already considering it, please think about moving
the documentation masters of your project to DocBook (or some format
from which you can generate DocBook). If everybody moved to using
DocBook as a common exchange format, it would become much easier to
support unified browsing of all system documentation with Web-like
hypertext capabilities, automatic indexing, and rich search
facilities.</p>

<p>Tools to generate man pages, HTML, and PostScript from DocBook
files are open-source and generally available.  My program, doclifter,
should make moving your manual-page masters to DocBook a fairly
painless process.</p>

<p>Many major open source projects (including the Linux kernel, the
Linux Documentation Project, X.org, GNOME, KDE, and FreeBSD) have
already moved to DocBook or are in the process of doing so.</p>

<p>(Individual entries for accepted patches are no longer shown.) </p>

<p>Summary: %d patches pending, %d accepted, %d rejected.</p>

<p>Status codes are as follows:</p>

<br />

<table width='100%%' border='1'>
<tr><td>n</td>
<td>No response yet.</td>
</tr>

<tr><td>p</td>
<td>Maintainer has informed me that this is fixed in the masters, but
I have not seen the fix yet.</td>
</tr>

<tr><td>y</td>
<td>Accepted</td>
</tr>

<tr><td>r</td>
<td>Rejected</td>
</tr>

<tr><td>s</td>
<td>Superseded (page lifts correctly without the patch)</td>
</tr>

<tr><td>[0-9]+</td>
<td>number of mailings sent</td>
</tr>

<tr><td>b</td>
<td>Address is blocked</td>
</tr>
</table>

<br />

<p>Problem codes are explained after the table.</p>

<br clear='left'/>

<table width='100%%' border='1'>
<tr><td><b>Patch:</b></td><td><b>Problem code:</b></td><td>Status:</td></tr>
''' % (time.strftime("%d %b %Y", time.gmtime()),
       buglist.distro, total, already,
       ok, percent(ok),
       total - already,
       100.00 - percent(ok),
       applied, percent(applied),
       errors, percent(errors),
       pending, accepted, rejected)

for line in bugs.lines:
    (status, pages, problems, mailto) = map(string.strip, line.split("|"))
    if 'g' in status:
        continue
    mailto = mailto.replace("<", "&lt;").replace(">", "&gt;").replace("@", "&#x40;")
    problems = " ".join(map(lambda x: "<a href='#%s'>%s</a>" % (x,x), problems))
    pages = "<br />\n".join(map(lambda x: "<a href='prepatch/%s.patch'>%s</a>" % (x,x), pages.split(",")))+"<br />\n"
    if 'p' in status or 'y' in status or 's' in status:
        status = "<font color='green'>" + status + "</font>"
    print "<tr><td>%s</td><td>%s</td><td>%s</td></tr>" % (pages,problems,status)
print "</table>\n\n<h1>Error codes:</h1>\n<dl>\n"

items = bugs.codes.items()
items.sort()
for (key, value) in items:
    print "<dt><a name='%s'>%s</a></dt>" % (key, key)
    print "<dd>%s</dd>" % value

print '''
</dl>

</div>
<hr />
</body>
</html>

<!--
Local Variables:
compile-command: "(cd ~/WWW; upload doclifter/problems.html)"
End:
-->
'''




#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: Copyright Eric S. Raymond <esr@thyrsus.com>
# SPDX-License-Identifier: BSD-2-Clause

# pylint: disable=line-too-long,missing-module-docstring,multiple-imports,invalid-name,redefined-outer-name,missing-class-docstring,too-many-instance-attributes,missing-function-docstring,unspecified-encoding,consider-using-with,consider-using-f-string,too-many-branches,too-many-statements,too-few-public-methods

import os, os.path, sys

subdir = "prepatch"

distro = "Xubuntu 18.04 with some extras"
future = "next Xubuntu release"

CATEGORY_COUNT = 8


class Buglist:
    def __init__(self, filename="PATCHES"):
        with open(filename) as file:
            self.codes = {}
            while True:
                line = file.readline()
                if line == "%%\n":
                    break
                if line[0] == "#":
                    continue
                if not line[0].isspace():
                    key = line[0]
                    self.codes[key] = ""
                    line = line[1:]
                self.codes[key] += line.strip() + "\n"

            self.maildict = {}
            self.lines = []
            self.fields = []
            self.typeseen = {}
            self.promised = 0
            self.counts = [0] * CATEGORY_COUNT
            self.patched = 0
            self.warnings = 0
            while True:
                line = file.readline()
                if not line:
                    break
                self.lines.append(line)
                fields = line.split("|")
                try:
                    (status, pages, problems, mailto) = list(
                        map(lambda x: x.strip(), fields)
                    )
                    if "p" in status:
                        self.promised += len(pages.split(","))
                except ValueError:
                    print("Wrong field count")
                    print(line)
                    sys.exit(1)
                for c in problems:
                    self.typeseen[c] = True
                self.fields.append(list(map(lambda x: x.strip(), fields)))
                if mailto not in self.maildict:
                    self.maildict[mailto] = []
                self.maildict[mailto].append((status, pages, problems))
        try:
            with open("full.log") as lf:
                warn_latch = False
                while True:
                    line = lf.readline()
                    if not line:
                        break
                    if not line.strip():
                        if warn_latch:
                            self.warnings += 1
                        continue
                    if "warning -" in line:
                        warn_latch = True
                    if line[0] != "!":
                        continue
                    warn_latch = False
                    line = line[2:]
                    rcolon = line.rindex("=")
                    retval = line[rcolon + 1 :].split()[0]
                    if retval.endswith("*"):
                        self.patched += 1
                        retval = retval[:-1]
                    self.counts[int(retval)] += 1
        except (OSError, IOError):
            sys.stderr.write("Conversion log is missing.\n")

    def pagelist(self, include="", exclude=""):
        lst = []
        for (status, pages, _, _) in self.fields:
            addit = not include
            for c in include:
                if c in status:
                    addit = True
                    break
            for c in exclude:
                if c in status:
                    addit = False
                    break
            if addit:
                lst += list(map(lambda x: x.strip(), pages.split(",")))
        lst.sort()
        return lst


def filestem(x):
    if x.endswith(".patch"):
        return x[:-6]
    if x.endswith(".correction"):
        return x[:-11]
    return x


def pagetofile(page):
    page = os.path.join(subdir, page)
    if os.path.exists(page + ".patch"):
        fp = open(page + ".patch")
        txt = fp.read()
        fp.close()
    # manlifter doesn't pick up corrections
    elif os.path.exists(page + ".correction"):
        fp = open(page + ".correction")
        txt = fp.read()
        fp.close()
    else:
        txt = None
    return txt


if __name__ == "__main__":
    import getopt

    (options, arguments) = getopt.getopt(sys.argv[1:], "e:")
    error_type_filter = None
    for (opt, val) in options:
        if opt == "-e":
            error_type_filter = val

    bugs = Buglist()
    if error_type_filter:  # report on what entries contain specified code
        for (status, pages, problems, mailto) in bugs.fields:
            if error_type_filter in problems:
                print("|".join((status, pages, problems, mailto)))
    else:
        # Default action is to sanity-check the database
        patches = set(filestem(x) for x in os.listdir(subdir))
        pages = set(bugs.pagelist())
        unresolved = set(bugs.pagelist(exclude="y"))
        resolved = set(bugs.pagelist(include="y"))
        counts = {}
        for page in pages:
            counts[page] = counts.get(page, 0) + 1
        duplicates = [x for x in pages if counts[x] > 1]
        if duplicates:
            print("Duplicates:", duplicates)
        print(
            "%d unresolved patches, %d resolved, %d promised, %d total, %d patches."
            % (len(unresolved), len(resolved), bugs.promised, len(pages), len(patches))
        )
        if patches - pages:
            print("These patches have no buglist entry:", patches - pages)
        if unresolved - patches:
            print("These bugs have no filelist entry:", unresolved - patches)
        if len(patches - unresolved):
            print("Leftovers:", " ".join(patches - unresolved))
        nonbugs = []
        for c in list(bugs.codes.keys()):
            if not c in bugs.typeseen:
                nonbugs.append(c)
        if nonbugs:
            print("These bug types no longer occur: ", ", ".join(nonbugs))
        available = list("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")
        for c in list(bugs.codes.keys()):
            available.remove(c)
        if available:
            print("These bug keys are available: ", ", ".join(available))
        else:
            print("All bug keys are in use")

# end



#!/usr/bin/env python3
import string, buglist, smtplib, time, hashlib
import email.mime.multipart, email.mime.text

def send(contact, msg):
    "Send a message."
    try:
        server = smtplib.SMTP('localhost')
        #server.set_debuglevel(1)
        server.sendmail("esr@thyrsus.com", [contact], msg)
        server.quit()
        print contact, "OK"
    except smtplib.SMTPServerDisconnected:
        print "*** Disconnected while mailing", contacr
    except smtplib.SMTPSenderRefused:
        print "*** Host refused sender for", contact
    except smtplib.SMTPRecipientsRefused:
        print "*** Recipient address refused for", contact
    except smtplib.SMTPDataError:
        print "*** Message data refused for", contact
    except smtplib.SMTPConnectError:
        print "*** Connection refused for", contact
    except smtplib.SMTPHeloError:
        print "*** HELO refused for", contact

def hashkey(patch):
    body = "\n".join(patch.split('\n')[2:])	# Trim off the header lines
    m = hashlib.md5()
    m.update(body)
    return m.digest()

def mailall(filter, hook):
    duplicates = {}
    for contact in bugs.maildict:
        if not contact or contact.startswith("http"):
            continue
        template = '''\
This is automatically generated email about markup problems in a man
page for which you appear to be responsible.  If you are not the right
person or list, please tell me so I can correct my database.

See http://catb.org/~esr/doclifter/bugs.html for details on how and
why these patches were generated.  Feel free to email me with any
questions.  Note: These patches do not change the modification date of
any manual page.  You may wish to do that by hand.

I apologize if this message seems spammy or impersonal. The volume of
markup bugs I am tracking is over five hundred - there is no real
alternative to generating bugmail from a database and template.
'''
        reminder = '''
My records indicate that you have accepted this patch, so this is just
a reminder.
'''
        unique = {}
        pagelist = []
        page_to_problems = {}
        for (status, pages, problems) in bugs.maildict[contact]:
            if 'y' in status or 'r' in status or 'b' in status or 's' in status:
                continue
            if filter(status, pages, problems, contact):
                # Setup
                for page in pages.split(","):
                    pagelist.append(page)
                    duplicates[page] = []
                    page_to_problems[page] = problems
        # Enable duplicate detection
        for page in pagelist:
            patch = buglist.pagetofile(page.strip())
            if patch is None:
                # Meant to generate a unique cookie
                body = str(time.time())
            else:
                d = hashkey(patch)
                if d in unique:
                    if not d in duplicates:
                        duplicates[d] = []
                    duplicates[d].append(page)
                else:
                    unique[d] = page
        # Now generate explanations
        explanations = []
        for page in unique.values():
            patch = buglist.pagetofile(page.strip())
            d = hashkey(patch)
            explanation =  "Problems with " + page + ":\n"
            if d in duplicates:
                explanation += "\n(Identical patches should apply to: %s)\n" \
                               % (" ".join(duplicates[d],))
            if 'p' in status:
                explanation += reminder
            if page.endswith("pm"):
                explanation += "\n(May reflect bugs in POD).\n"
            explanation += "\n"
            patch = buglist.pagetofile(page.strip())
            for letter in page_to_problems[page]:
                if letter.isalnum():
                    explanation += "%s\n" % (bugs.codes[letter])
                elif letter != '*':
                    sys.stderr.write("Problem with %s\n" % page)
            d = hashkey(patch)
            if patch:
                this =  explanation + patch
            else:
                this += explanation + "(No patch.)\n"
            explanations.append(this)
        if explanations:
            pagelist = ", ".join(pagelist)
            if len(pagelist) > 100:
                pagelist = "several man pages you maintain"
            body = template \
                  + "\n--\n                             Eric S. Raymond\n"
            msg = email.mime.multipart.MIMEMultipart()
            msg["To"] = contact
            msg["Subject"] = "Problems in " + pagelist
            msg.attach(email.mime.text.MIMEText(body))
            for x in explanations:
                msg.attach(email.mime.text.MIMEText(x))
            apply(hook, (contact, str(msg) + "\n"))

if __name__ == '__main__':
    import getopt, re, sys

    bugs = buglist.Buglist()
    selector = None
    pagefilter = None
    codefilter = None
    addressfilter = None
    hook = lambda contact, msg: sys.stdout.write(msg)
    (options, arguments) = getopt.getopt(sys.argv[1:], "a:c:mp:s:")
    for (switch, val) in options:
        if switch == '-s':	# Select on given status code
            selector = val
        elif switch == '-p':	# Filter on given regexp matching pages
            pagefilter = val
        elif switch == '-a':	# Filter on given regexp matching addressees
            addressfilter = val
        elif switch == '-c':	# Filter on given regexp matching problems
            codefilter = val
        elif switch == '-m':	# Actually send mail
            hook = send
    if selector:
        selector = re.compile(selector)
    if pagefilter:
        pagefilter = re.compile(pagefilter)
    if addressfilter:
        addressfilter = re.compile(addressfilter)
    if codefilter:
        codefilter = re.compile(codefilter)

    def filter(status, pages, problems, contact):
        if selector and not selector.search(status):
            return False
        if pagefilter and not pagefilter.search(pages):
            return False
        if addressfilter and not addressfilter.search(contact):
            return False
        if codefilter and not codefilter.search(contact):
            return False
        return True

    mailall(filter, hook)
